from typing import Any, Union

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519
from cryptography.x509 import Certificate
from cryptography.x509.oid import NameOID
from pycose.keys import CoseKey


def selfsigned_x509cert(
    cert_info: dict[str, Any],
    private_key: CoseKey,
    encoding: str = "DER"
) -> Union[Certificate, bytes]:
    """
    Returns an X.509 certificate derived from the private key of the MSO Issuer

    :param cert_info: dict[str, Any]: the certificate information, should contain
        at least one of: country_name, state_or_province_name, locality_name,
        organization_name, common_name, not_valid_before, not_valid_after, san_url
    :param private_key: CoseKey: the private key to use for signing the certificate
    :param encoding: str: the encoding to use, default is DER
    :return: Union[Certificate, bytes]: the X.509 certificate
    """
    if not private_key:
        raise ValueError("private_key must be set")

    # convert the private key to a cryptography private key instance
    if hasattr(private_key, "kty") and private_key.kty is not None and hasattr(private_key.kty, "identifier"):
        if private_key.kty.identifier == 2:  # EC2Key
            private_key_inst = ec.derive_private_key(
                int.from_bytes(private_key['d'], byteorder="big"), ec.SECP256R1()
            )
        elif private_key.kty.identifier == 1:  # OKPKey
            private_key_inst = ed25519.Ed25519PrivateKey.from_private_bytes(
                private_key['d']
            )
        else:
            raise ValueError(f"Unsupported key type: {private_key.kty}")
    else:
        raise ValueError(
            "private_key.kty or private_key.kty.identifier is not set or unknown"
        )

    public_key_inst = private_key_inst.public_key()

    name_attributes = []
    if "country_name" in cert_info:
        name_attributes.append(x509.NameAttribute(NameOID.COUNTRY_NAME, cert_info["country_name"]))
    if "state_or_province_name" in cert_info:
        name_attributes.append(x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, cert_info["state_or_province_name"]))
    if "locality_name" in cert_info:
        name_attributes.append(x509.NameAttribute(NameOID.LOCALITY_NAME, cert_info["locality_name"]))
    if "organization_name" in cert_info:
        name_attributes.append(x509.NameAttribute(NameOID.ORGANIZATION_NAME, cert_info["organization_name"]))
    if "common_name" in cert_info:
        name_attributes.append(x509.NameAttribute(NameOID.COMMON_NAME, cert_info["common_name"]))

    subject = issuer = x509.Name(name_attributes)

    cert_builder = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).public_key(
        public_key_inst
    ).serial_number(
        x509.random_serial_number()
    )

    if "not_valid_before" in cert_info:
        cert_builder = cert_builder.not_valid_before(
            cert_info["not_valid_before"]
        )

    if "not_valid_after" in cert_info:
        cert_builder = cert_builder.not_valid_after(
            cert_info["not_valid_after"]
        )

    if "san_url" in cert_info:
        cert_builder = cert_builder.add_extension(
            x509.SubjectAlternativeName(
                [
                    x509.UniformResourceIdentifier(
                        cert_info["san_url"]
                    )
                ]
            ),
            critical=False,
            # Sign our certificate with our private key
        )

    cert = cert_builder.sign(private_key_inst, hashes.SHA256())

    if not encoding:
        return cert
    else:
        return cert.public_bytes(
            getattr(serialization.Encoding, encoding)
        )



"""
Test support for the 'errors' field in MobileDocument.

ISO 18013-5 specifies that when status != 0, documents may contain
an 'errors' field describing which elements were not available.
"""

from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mdoc.verifier import MobileDocument
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY


def test_mobile_document_with_errors_field():
    """Test that MobileDocument accepts an 'errors' field."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]

    # Add errors field (simulating status 20 - elements not present)
    document['errors'] = {
        'org.micov.medical.1': {
            'missing_element': 1  # Error code for element not present
        }
    }

    # Should not raise TypeError
    doc = MobileDocument(**document)

    assert doc.doctype == "org.micov.medical.1"
    assert doc.errors is not None
    assert isinstance(doc.errors, dict)


def test_mobile_document_without_errors_field():
    """Test that MobileDocument works without 'errors' field (backward compatibility)."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]

    # No errors field
    doc = MobileDocument(**document)

    assert doc.doctype == "org.micov.medical.1"
    assert doc.errors == {}  # Should default to empty dict


def test_mobile_document_dump_with_errors():
    """Test that dump() includes errors field when present."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]

    # Add errors field
    errors_data = {
        'org.micov.medical.1': {
            'missing_element': 1
        }
    }
    document['errors'] = errors_data

    doc = MobileDocument(**document)
    dump = doc.dump()

    assert dump
    assert isinstance(dump, bytes)

    # Decode and verify errors field is present
    import cbor2
    decoded = cbor2.loads(dump)
    # The dump is wrapped in a CBORTag, so we need to access .value
    if hasattr(decoded, 'value'):
        decoded = decoded.value

    assert 'errors' in decoded
    assert decoded['errors'] == errors_data


def test_mobile_document_dump_without_errors():
    """Test that dump() works without errors field (backward compatibility)."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    dump = doc.dump()

    assert dump
    assert isinstance(dump, bytes)

    # Decode and verify errors field is NOT present
    import cbor2
    decoded = cbor2.loads(dump)
    if hasattr(decoded, 'value'):
        decoded = decoded.value

    # errors field should not be in dump if it's empty
    assert 'errors' not in decoded



import json
import random

import cbor2
from cbor2.tool import DefaultEncoder, key_to_str
from pycose.messages import Sign1Message


def bytes2CoseSign1(data: bytes) -> Sign1Message:
    """
    Gets bytes and return a COSE_Sign1 object

    :param data: bytes: the COSE Sign1 as bytes
    :return: Sign1Message: the COSE Sign1 object
    """
    decoded = Sign1Message.decode(cbor2.loads(data).value)

    return decoded


def cborlist2CoseSign1(data: list) -> Sign1Message:
    """
    Gets cbor2 decoded COSE Sign1 as a list and return a COSE_Sign1 object

    :param data: list: the COSE Sign1 as a list
    :return: Sign1Message: the COSE Sign1 object
    """
    decoded = Sign1Message.decode(
        cbor2.dumps(
            cbor2.CBORTag(18, value=data)
        )
    )

    return decoded


def pretty_print(cbor_loaded: dict) -> None:
    """"
    Pretty print a CBOR object

    :param cbor_loaded: dict: the CBOR object
    """
    _obj = key_to_str(cbor_loaded)
    res = json.dumps(
        _obj,
        indent=(None, 4),
        cls=DefaultEncoder
    )
    print(res)


def shuffle_dict(d: dict) -> dict:
    """
    Shuffle a dictionary

    :param d: dict: the dictionary to shuffle
    :return: dict: the shuffled dictionary
    """

    keys = list(d.keys())

    for i in range(random.randint(3, 27)):  # nosec: B311
        random.shuffle(keys)

    return dict([(key, d[key]) for key in keys])



from typing import Union

import cbor2

from pymdoccbor.mso.verifier import MsoVerifier


class IssuerSigned:
    """
    nameSpaces provides the definition within which the data elements of
        the document are defined.
        A document may have multiple nameSpaces.

    IssuerAuth is a COSE_Sign1 ; The payload is the MobileSecurityObject,
        see ISO 18013-5 section 9.2.2.4

    issuerAuth is a list of [
        cbor({1: -7}) # Protected Header, find -7
            here https://datatracker.ietf.org/doc/html/rfc8152
        cbor({33: bytes}) # Unprotected Header containing X509 certificate
        cbor({24: bytes}) # Payload -> Mobile Security Object
        bytes # Signature
    ]
    """

    def __init__(self, nameSpaces: dict, issuerAuth: Union[cbor2.CBORTag, dict, bytes]) -> None:
        """
        Initialize the IssuerSigned object

        :param nameSpaces: dict: the nameSpaces of the document
        :param issuerAuth: Union[dict, bytes]: the issuerAuth info of the document
        """

        self.namespaces: dict = nameSpaces

        if not issuerAuth:
            raise ValueError("issuerAuth must be provided")

        #  if isinstance(ia, dict):
        self.issuer_auth = MsoVerifier(issuerAuth)

    def dump(self) -> dict:
        """
        It returns the issuerSigned as a dict

        :return: dict: the issuerSigned as a dict
        """

        return {
            'nameSpaces': self.namespaces,
            'issuerAuth': self.issuer_auth
        }

    def dumps(self) -> bytes:
        """
        It returns the issuerSigned as bytes

        :return: dict: the issuerSigned as bytes
        """

        return cbor2.dumps(
            {
                'nameSpaces': self.namespaces,
                'issuerAuth': self.issuer_auth.payload_as_cbor
            }
        )



from datetime import datetime, timedelta, timezone

CERT_DATA = {
    "country_name": "US",
    "state_or_province_name": "California",
    "locality_name": "San Francisco",
    "organization_name": "Micov",
    "common_name": "My Company",
    "san_url": "https://credential-issuer.example.org",
    "not_valid_before": datetime.now(timezone.utc) - timedelta(days=1),
    "not_valid_after": datetime.now(timezone.utc) + timedelta(days=10),
}



from pycose.messages import CoseMessage

from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mso.verifier import MsoVerifier
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY

mdoc = MdocCborIssuer(
    private_key=PKEY,
    alg="ES256",
    cert_info=CERT_DATA
)

mdoc.new(
    data=MICOV_DATA,
    # devicekeyinfo=PKEY,  # TODO
    doctype="org.micov.medical.1",
    validity={
        "issuance_date": "2024-12-31",
        "expiry_date": "2050-12-31"
    },
)


def test_mso_verifier_fail():
    try:
        MsoVerifier(None)
    except Exception as e:
        assert str(e) == "MsoParser only supports raw bytes and list, a <class 'NoneType'> was provided"


def test_mso_verifier_creation():
    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]

    msov = MsoVerifier(issuerAuth)

    assert isinstance(msov.object, CoseMessage)


def test_mso_verifier_verify_signatures():
    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]

    msov = MsoVerifier(issuerAuth)

    assert msov.verify_signature()


def test_mso_verifier_payload_as_cbor():
    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]

    msov = MsoVerifier(issuerAuth)

    cbor = msov.payload_as_dict

    assert cbor
    assert cbor["version"] == "1.0"
    assert cbor["digestAlgorithm"] == "SHA-256"
    assert cbor["valueDigests"]["org.micov.medical.1"]


def test_payload_as_raw():
    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]

    msov = MsoVerifier(issuerAuth)

    raw = msov.payload_as_raw

    assert raw
    assert isinstance(raw, bytes)
    assert len(raw) > 0



from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mdoc.issuersigned import IssuerSigned
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY

mdoc = MdocCborIssuer(
    private_key=PKEY,
    alg="ES256",
    cert_info=CERT_DATA
)
mdoc.new(
    data=MICOV_DATA,
    # devicekeyinfo=PKEY,  # TODO
    doctype="org.micov.medical.1",
    validity={
        "issuance_date": "2024-12-31",
        "expiry_date": "2050-12-31"
    },
)
issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]
issuer_signed = IssuerSigned(**issuerAuth)


def test_issuer_signed_fail():
    try:
        IssuerSigned(None, None)
    except Exception as e:
        assert str(e) == "issuerAuth must be provided"


def test_issuer_signed_creation():
    assert issuer_signed.namespaces
    assert issuer_signed.issuer_auth


def test_issuer_signed_dump():
    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]

    issuer_signed = IssuerSigned(**issuerAuth)

    dump = issuer_signed.dump()
    assert dump
    assert dump["nameSpaces"] == issuer_signed.namespaces
    assert dump["issuerAuth"] == issuer_signed.issuer_auth


def test_issuer_signed_dumps():
    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]

    issuer_signed = IssuerSigned(**issuerAuth)

    dumps = issuer_signed.dumps()
    assert dumps
    assert isinstance(dumps, bytes)
    assert len(dumps) > 0



MICOV_DATA = {
    "org.micov.medical.1": {
        "last_name": "Rossi",
        "given_name": "Mario",
        "birth_date": "1922-03-13",
        "PersonId_nic": {
            "PersonIdNumber": "1234567890",
            "PersonIdType": "nic",
            "PersonIdIS": "IT",
        },
        "sex": 1,
        "VPInfo_COVID-19_1": {
            "VaccineProphylaxis": "",
            "VaccMedicinalProd": "Moderna",
            "VaccMktAuthHolder": "Moderna",
            "VaccDoseNumber": "2/2",
            "VaccAdmDate": "2021-01-01",
            "VaccCountry": "IT",
        },
        "CertIssuer": "Italian Ministry of Health",
        "CertId": "1234567890",
    }
}



"""
Test certificate chain verification and element hash verification.
"""

from datetime import datetime, timedelta, timezone

import cbor2
import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509.oid import NameOID

from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mdoc.verifier import MdocCbor, MobileDocument
from pymdoccbor.mso.verifier import MsoVerifier
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY


def generate_test_certificates():
    """Generate a test root CA and DS certificate for testing."""
    # Generate root CA private key
    root_key = ec.generate_private_key(ec.SECP256R1(), default_backend())

    # Create root CA certificate
    root_subject = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Test Root CA"),
        x509.NameAttribute(NameOID.COMMON_NAME, "Test Root CA"),
    ])

    root_cert = (
        x509.CertificateBuilder()
        .subject_name(root_subject)
        .issuer_name(root_subject)
        .public_key(root_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=365))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .sign(root_key, hashes.SHA256(), default_backend())
    )

    # Generate DS private key
    ds_key = ec.generate_private_key(ec.SECP256R1(), default_backend())

    # Create DS certificate signed by root CA
    ds_subject = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Test DS"),
        x509.NameAttribute(NameOID.COMMON_NAME, "Test Document Signer"),
    ])

    ds_cert = (
        x509.CertificateBuilder()
        .subject_name(ds_subject)
        .issuer_name(root_subject)
        .public_key(ds_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=365))
        .sign(root_key, hashes.SHA256(), default_backend())
    )

    return root_cert, ds_cert, ds_key


def test_certificate_chain_verification_success():
    """Test successful certificate chain verification."""
    root_cert, ds_cert, ds_key = generate_test_certificates()

    # Create mdoc with DS certificate
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]
    msov = MsoVerifier(issuerAuth)

    # Replace the certificate in the MSO with our test DS cert
    msov.x509_certificates = [ds_cert]

    # Verify with trusted root
    verified_root = msov.attest_public_key([root_cert])

    assert verified_root is not None
    assert verified_root == root_cert


def test_certificate_chain_verification_untrusted():
    """Test certificate chain verification with untrusted root."""
    root_cert, ds_cert, ds_key = generate_test_certificates()

    # Generate a different root that didn't sign the DS cert
    untrusted_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
    untrusted_subject = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Untrusted Root"),
        x509.NameAttribute(NameOID.COMMON_NAME, "Untrusted Root CA"),
    ])
    untrusted_cert = (
        x509.CertificateBuilder()
        .subject_name(untrusted_subject)
        .issuer_name(untrusted_subject)
        .public_key(untrusted_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=365))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .sign(untrusted_key, hashes.SHA256(), default_backend())
    )

    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]
    msov = MsoVerifier(issuerAuth)
    msov.x509_certificates = [ds_cert]

    # Should raise ValueError
    with pytest.raises(ValueError, match="not signed by any trusted root"):
        msov.attest_public_key([untrusted_cert])


def test_certificate_chain_verification_skipped():
    """Test that verification is skipped when no trusted roots provided."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]
    msov = MsoVerifier(issuerAuth)

    # Should return None and log warning
    result = msov.attest_public_key(None)
    assert result is None


def test_element_hash_verification_success():
    """Test successful element hash verification."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]
    namespaces = mdoc.signed["documents"][0]["issuerSigned"]["nameSpaces"]

    msov = MsoVerifier(issuerAuth)
    results = msov.verify_element_hashes(namespaces)

    assert results['valid'] is True
    assert results['total'] > 0
    assert results['verified'] == results['total']
    assert len(results['failed']) == 0


def test_element_hash_verification_tampered():
    """Test element hash verification with tampered data."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    issuerAuth = mdoc.signed["documents"][0]["issuerSigned"]["issuerAuth"]
    namespaces = mdoc.signed["documents"][0]["issuerSigned"]["nameSpaces"]

    # Tamper with an element
    namespace_key = list(namespaces.keys())[0]
    if namespaces[namespace_key]:
        first_item = namespaces[namespace_key][0]
        # first_item is already a CBORTag object
        if isinstance(first_item, cbor2.CBORTag):
            item_content = cbor2.loads(first_item.value)
            item_content['elementValue'] = 'TAMPERED'
            # Re-encode
            namespaces[namespace_key][0] = cbor2.CBORTag(24, cbor2.dumps(item_content))

    msov = MsoVerifier(issuerAuth)
    results = msov.verify_element_hashes(namespaces)

    assert results['valid'] is False
    assert results['total'] > 0
    assert results['verified'] < results['total']
    assert len(results['failed']) > 0
    assert results['failed'][0]['reason'] == 'hash mismatch'


def test_mobile_document_verify_with_hashes():
    """Test MobileDocument.verify() with hash verification enabled."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    # Verify with hash verification enabled
    is_valid = doc.verify(verify_hashes=True)

    assert is_valid is True
    assert doc.hash_verification is not None
    assert doc.hash_verification['valid'] is True
    assert doc.hash_verification['total'] > 0


def test_mobile_document_verify_without_hashes():
    """Test MobileDocument.verify() with hash verification disabled."""
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    # Verify with hash verification disabled
    is_valid = doc.verify(verify_hashes=False)

    assert is_valid is True
    assert doc.hash_verification is None


def test_mdoc_cbor_verify_with_all_features():
    """Test MdocCbor.verify() with certificate chain and hash verification."""
    root_cert, ds_cert, ds_key = generate_test_certificates()

    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    # Use the full signed structure, not just dumps()
    mdoc_cbor = MdocCbor()
    mdoc_cbor.loads(cbor2.dumps(mdoc.signed))

    # Verify with hash verification (cert chain will be skipped without trusted roots)
    is_valid = mdoc_cbor.verify(verify_hashes=True)

    assert is_valid is True
    assert len(mdoc_cbor.documents) > 0
    assert mdoc_cbor.documents[0].hash_verification is not None
    assert mdoc_cbor.documents[0].hash_verification['valid'] is True



from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mdoc.verifier import MobileDocument
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY


def test_verifier_must_fail_document_type():
    try:
        MobileDocument(None, None)
    except Exception as e:
        assert str(e) == "You must provide a document type"


def test_verifier_must_fail_issuer_signed():
    try:
        MobileDocument("org.micov.medical.1", None)
    except Exception as e:
        assert str(e) == "You must provide a signed document"


def test_mobile_document():
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        # devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    assert doc.doctype == "org.micov.medical.1"
    assert doc.issuersigned


def test_mobile_document_dump():
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        # devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    dump = doc.dump()
    assert dump
    assert isinstance(dump, bytes)
    assert len(dump) > 0


def test_mobile_document_dumps():
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        # devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    dumps = doc.dumps()
    assert dumps
    assert isinstance(dumps, bytes)
    assert len(dumps) > 0


def test_mobile_document_verify():
    mdoc = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    mdoc.new(
        data=MICOV_DATA,
        # devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    document = mdoc.signed["documents"][0]
    doc = MobileDocument(**document)

    assert doc.verify()



import base64

PKEY = {
    'KTY': 'EC2',
    'CURVE': 'P_256',
    'ALG': 'ES256',
    'D': b"<\xe5\xbc;\x08\xadF\x1d\xc5\x0czR'T&\xbb\x91\xac\x84\xdc\x9ce\xbf\x0b,\x00\xcb\xdd\xbf\xec\xa2\xa5",
    'KID': b"demo-kid"
}


def base64_urldecode(v: str) -> bytes:
    """Urlsafe base64 decoding. This function will handle missing
    padding symbols.

    :returns: the decoded data in bytes, format, convert to str use method '.decode("utf-8")' on result
    :rtype: bytes
    """
    padded = f"{v}{'=' * divmod(len(v), 4)[1]}"
    return base64.urlsafe_b64decode(padded)


decoded_x = base64_urldecode("dGLQBwQIPWjc2aA6zRc06wlNVxiw72PMwJlEXHEvP-E")
decoded_d = base64_urldecode("NOHGihpyjNa_xBSd17Wr4ynkSM-afunMgpoPoFkelhI")

PKEY_ED25519 = {
    'KTY': 'OKP',
    'CURVE': 'Ed25519',
    'ALG': 'EdDSA',
    'D': decoded_d,
    'X': decoded_x,
    'KID': b"demo-kid-ed25519"
}

PKEY_RSA = {
    'KTY': 'RSA',
    'E': b'\xd4\xf1\xf2o',
    'N': b'[_\x81\\6y3\xbf\x01\xad\xba\xe26&\xcb\xa2g\xff\x97\xa1rv\xa7\x9a{\xfb\x01r^S\xfb\xefY\xb4\x14\xcesz\x99H\x02\xaf\xf5\xab\x18_\xac\xaaR\x13Q\xe6\xa0\x9a\x8c\x8a\x1f\x13\x0b9\xf3\xbb\xe1\x0b\xb9<\xe7\xc0\xffU\xa0\xcb\x1aw\xf2/\x11\x0e\xea^\x98:cp\x1f3\xc9\x81\x93e\x81\xb4\xb20s\xa6\xaaV\xf3\x03y\xb3\xd9\x93i\x14\xa7\xafi.\x08\xdey\x15s-V\x10\xf0\x0f+:E\x10\xec\xca\x93\x17\xecg\xaf!\x11\xe7\x91\xcdG7)\x83\xc3\xdd\xc2xp\xb2v_\xf2l\xc9\xc7\x15r\xf9\xa1U\xe9`\xde\xf1\xa9\xc2\xb6\xde\xebc|\xef\xb0s,\x10\xa1l\x81&\xcc\xb9\xfa\xb6\xffs\x1a9\x0c[7\xafJ\x1c\xd5\xb6\xc7?\x1c\x8fN\x1a\xde\x7f\xa4\x8f\xf6,\xed\x89b\x87\xcaXL\x8e}\xa5K\x0b\x9a\x8c\xb2\xd2\x91\x0f\xedI\x8e\x8fYq#\x8c\xd1\x02\xe2B\xff\xf1\x1dT\x15\xb1I\xe8\xd8\xfc[\xd5Y\x9ab\xcc\xe3\xff\xac\xfa\x85',  # noqa: E501
    'KEY_OPS': ['SIGN']
}



class MissingPrivateKey(Exception):
    pass


class NoDocumentTypeProvided(Exception):
    pass


class NoSignedDocumentProvided(Exception):
    pass


class MissingIssuerAuth(Exception):
    pass


class InvalidStatusDescriptor(Exception):
    pass



import binascii
import logging
from typing import List

import cbor2

from pymdoccbor.exceptions import InvalidMdoc
from pymdoccbor.mdoc.exceptions import (NoDocumentTypeProvided,
                                        NoSignedDocumentProvided)
from pymdoccbor.mdoc.issuersigned import IssuerSigned

logger = logging.getLogger('pymdoccbor')


class MobileDocument:
    """
    MobileDocument class to handle the Mobile Document
    """

    _states = {
        True: "valid",
        False: "failed",
    }

    def __init__(self, docType: str, issuerSigned: dict, deviceSigned: dict = {}, errors: dict = None) -> None:
        """
        Initialize the MobileDocument object

        :param docType: str: the document type
        :param issuerSigned: dict: the issuerSigned info
        :param deviceSigned: dict: the deviceSigned info
        :param errors: dict: optional errors field (ISO 18013-5 status != 0)
        """

        if not docType:
            raise NoDocumentTypeProvided("You must provide a document type")

        self.doctype: str = docType  # eg: 'org.iso.18013.5.1.mDL'

        if not issuerSigned:
            raise NoSignedDocumentProvided("You must provide a signed document")

        self.issuersigned: List[IssuerSigned] = IssuerSigned(**issuerSigned)
        self.is_valid = False
        self.hash_verification = None  # Will store hash verification results
        self.devicesigned: dict = deviceSigned
        self.errors: dict = errors if errors is not None else {}

    def dumps(self) -> bytes:
        """
        It returns the AF binary repr as bytes

        :return: bytes: the document as bytes
        """
        return binascii.hexlify(self.dump())

    def dump(self) -> bytes:
        """
        It returns the document as bytes

        :return: dict: the document as bytes
        """
        doc_dict = {
            'docType': self.doctype,
            'issuerSigned': self.issuersigned.dumps()
        }

        # Include errors field if present (ISO 18013-5 status != 0)
        if self.errors:
            doc_dict['errors'] = self.errors

        return cbor2.dumps(
            cbor2.CBORTag(
                24,
                value=doc_dict
            )
        )

    def verify(self, trusted_root_certs: list = None, verify_hashes: bool = True) -> bool:
        """
        Verify the document signature and optionally element hashes

        Args:
            trusted_root_certs: List of trusted root certificates for chain validation
            verify_hashes: If True, also verify element hashes against MSO
        :return: bool: True if the signature is valid, False otherwise
        """
        # Verify signature
        self.is_valid = self.issuersigned.issuer_auth.verify_signature(trusted_root_certs)

        # Verify element hashes if requested
        if verify_hashes and self.is_valid:
            hash_results = self.issuersigned.issuer_auth.verify_element_hashes(
                self.issuersigned.namespaces
            )
            self.hash_verification = hash_results
            self.is_valid = self.is_valid and hash_results['valid']

        return self.is_valid

    def __repr__(self) -> str:
        return f"{self.__module__}.{self.__class__.__name__} [{self._states[self.is_valid]}]"


class MdocCbor:
    """
    MdocCbor class to handle the Mobile Document
    """

    def __init__(self) -> None:
        """
        Initialize the MdocCbor object
        """
        self.data_as_bytes: bytes = b""
        self.data_as_cbor_dict: dict = {}

        self.documents: List[MobileDocument] = []
        self.documents_invalid: list = []
        self.disclosure_map: dict = {}

    def loads(self, data: str) -> None:
        """
        Load the data from a AF Binary string

        :param data: str: the AF binary string
        """
        if isinstance(data, bytes):
            data = binascii.hexlify(data)

        self.data_as_bytes = binascii.unhexlify(data)
        self.data_as_cbor_dict = cbor2.loads(self.data_as_bytes)

    def dump(self) -> bytes:
        """
        Returns the CBOR representation of the mdoc as bytes
        """
        return self.data_as_bytes

    def dumps(self) -> bytes:
        """
        Returns the AF binary representation of the mdoc as bytes

        :return: bytes: the AF binary representation of the mdoc
        """
        return binascii.hexlify(self.data_as_bytes)

    @property
    def data_as_string(self) -> str:
        return self.dumps().decode()

    def _decode_claims(self, claims: list[dict]) -> dict:
        decoded_claims = {}

        for claim in claims:
            decoded = cbor2.loads(claim.value)

            if isinstance(decoded['elementValue'], cbor2.CBORTag):
                decoded_claims[decoded['elementIdentifier']] = decoded['elementValue'].value
            elif isinstance(decoded['elementValue'], list):
                claims_list = []

                for element in decoded['elementValue']:
                    # Handle simple values in lists (strings, numbers, etc.)
                    if not isinstance(element, dict):
                        claims_list.append(element)
                        continue

                    # Handle dict elements
                    claims_dict = {}
                    for key, value in element.items():
                        if isinstance(value, cbor2.CBORTag):
                            claims_dict[key] = value.value
                        else:
                            claims_dict[key] = value
                    claims_list.append(claims_dict)

                decoded_claims[decoded['elementIdentifier']] = claims_list
            else:
                decoded_claims[decoded['elementIdentifier']] = decoded['elementValue']

        return decoded_claims

    def verify(self, trusted_root_certs: list = None, verify_hashes: bool = True) -> bool:
        """
        Verify signatures of all documents contained in the mdoc

        Args:
            trusted_root_certs: List of trusted root certificates (x509.Certificate objects)
                               for chain validation. If None, skips chain validation.
            verify_hashes: If True, also verify element hashes against MSO valueDigests
        :return: bool: True if all signatures are valid, False otherwise
        """
        cdict = self.data_as_cbor_dict

        for i in ('version', 'documents'):
            if i not in cdict:
                raise InvalidMdoc(
                    f"Mdoc is invalid since it doesn't contain the '{i}' element"
                )

        doc_cnt = 1
        for doc in cdict['documents']:
            mso = MobileDocument(**doc)

            try:
                if mso.verify(trusted_root_certs, verify_hashes):
                    self.documents.append(mso)
                else:
                    self.documents_invalid.append(mso)

                for namespace, claims in mso.issuersigned.namespaces.items():
                    self.disclosure_map[namespace] = self._decode_claims(claims)

            except Exception as e:
                logger.error(
                    f"COSE Sign1 validation failed to the document number #{doc_cnt}. "
                    f"Then it is appended to self.documents_invalid: {e}"
                )
                self.documents_invalid.append(doc)

            doc_cnt += 1

        self.status = cdict.get('status', None)

        return False if self.documents_invalid else True

    def __repr__(self) -> str:
        return (
            f"{self.__module__}.{self.__class__.__name__} "
            f"[{len(self.documents)} valid documents]"
        )



class InvalidMdoc(Exception):
    """
    """


class UnsupportedMsoDataFormat(Exception):
    pass


class MsoPrivateKeyRequired(Exception):
    pass


class MsoX509ChainNotFound(Exception):
    pass



from pycose.messages import CoseMessage

from pymdoccbor.mso.issuer import MsoIssuer
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY


def test_mso_issuer_fail():
    try:
        MsoIssuer(None, None)
    except Exception as e:
        assert str(e) == "MSO Writer requires a valid private key"


def test_mso_issuer_creation():
    msoi = MsoIssuer(
        data=MICOV_DATA,
        private_key=PKEY,
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
        alg="ES256",
        cert_info=CERT_DATA
    )

    assert msoi.private_key
    assert msoi.data
    assert msoi.hash_map
    assert list(msoi.hash_map.keys())[0] == 'org.micov.medical.1'
    assert msoi.disclosure_map['org.micov.medical.1']


def test_mso_issuer_sign():
    msoi = MsoIssuer(
        data=MICOV_DATA,
        private_key=PKEY,
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
        alg="ES256",
        cert_info=CERT_DATA
    )

    mso = msoi.sign()
    assert isinstance(mso, CoseMessage)



import datetime
import hashlib
import logging
import secrets
import uuid
from typing import Union

import cbor2
from cbor_diag import cbor2diag
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.x509 import Certificate
from pycose.headers import Algorithm
from pycose.keys import CoseKey
from pycose.messages import Sign1Message

from pymdoccbor import settings
from pymdoccbor.exceptions import MsoPrivateKeyRequired
from pymdoccbor.tools import shuffle_dict
from pymdoccbor.x509 import selfsigned_x509cert

logger = logging.getLogger("pymdoccbor")


class MsoIssuer:
    """
    MsoIssuer helper class to create a new mso
    """

    def __init__(
        self,
        data: dict,
        validity: dict,
        cert_path: str | None = None,
        key_label: str | None = None,
        user_pin: str | None = None,
        lib_path: str | None = None,
        slot_id: int | None = None,
        kid: str | None = None,
        alg: str | None = None,
        hsm: bool | None = False,
        private_key: dict | CoseKey | None = None,
        digest_alg: str | None = settings.PYMDOC_HASHALG,
        revocation: dict | None = None,
        cert_info: dict | None = None,
    ) -> None:
        """
        Initialize a new MsoIssuer

        :param data: dict: the data to sign
        :param validity: validity: the validity info of the mso
        :param cert_path: str: the path to the certificate
        :param key_label: str: key label
        :param user_pin: str: user pin
        :param lib_path: str: path to the library cryptographic library
        :param slot_id: int: slot id
        :param kid: str: key id
        :param alg: str: hashig algorithm
        :param hsm: bool: hardware security module
        :param private_key: Union[dict, CoseKey]: the signing key
        :param digest_alg: str: the digest algorithm
        :param revocation: dict: revocation dict (may include status_list, identifier_list)
        """

        if private_key:
            if isinstance(private_key, dict):
                self.private_key = CoseKey.from_dict(private_key)
                if not self.private_key.kid:
                    self.private_key.kid = str(uuid.uuid4())
            elif isinstance(private_key, CoseKey):
                self.private_key = private_key
            else:
                raise ValueError("private_key must be a dict or CoseKey object")
        else:
            if not hsm:
                raise MsoPrivateKeyRequired("MSO Writer requires a valid private key")

        if not validity:
            raise ValueError("validity must be present")

        if not alg:
            raise ValueError("alg must be present")

        self.data: dict = data
        self.hash_map: dict = {}
        self.disclosure_map: dict = {}
        self.digest_alg = digest_alg
        self.key_label = key_label
        self.user_pin = user_pin
        self.lib_path = lib_path
        self.slot_id = slot_id
        self.hsm = hsm
        self.alg = alg
        self.kid = kid
        self.validity = validity
        self.revocation = revocation

        self.cert_path = cert_path
        self.cert_info = cert_info

        if not self.cert_path and (not self.cert_info or not self.private_key):
            raise ValueError(
                "cert_path or cert_info with a private key must be provided to properly insert a certificate"
            )

        alg_map = {"ES256": "sha256", "ES384": "sha384", "ES512": "sha512"}

        if self.alg not in alg_map:
            raise ValueError(f"Unsupported algorithm: {self.alg}")

        hashfunc = getattr(hashlib, alg_map[self.alg])

        digest_cnt = 0
        for ns, values in data.items():
            self.disclosure_map[ns] = {}
            self.hash_map[ns] = {}
            for k, v in shuffle_dict(values).items():
                _rnd_salt = secrets.token_bytes(settings.DIGEST_SALT_LENGTH)

                _value_cbortag = settings.CBORTAGS_ATTR_MAP.get(k, None)

                if _value_cbortag:
                    v = cbor2.CBORTag(_value_cbortag, value=v)

                if isinstance(v, dict):
                    for k2, v2 in v.items():
                        _value_cbortag = settings.CBORTAGS_ATTR_MAP.get(k2, None)
                        if _value_cbortag:
                            v[k2] = cbor2.CBORTag(_value_cbortag, value=v2)

                if isinstance(v, list) and k != "nationality":
                    for item in v:
                        for k2, v2 in item.items():
                            _value_cbortag = settings.CBORTAGS_ATTR_MAP.get(k2, None)
                            if _value_cbortag:
                                item[k2] = cbor2.CBORTag(_value_cbortag, value=v2)

                self.disclosure_map[ns][digest_cnt] = cbor2.CBORTag(
                    24,
                    value=cbor2.dumps(
                        {
                            "digestID": digest_cnt,
                            "random": _rnd_salt,
                            "elementIdentifier": k,
                            "elementValue": v,
                        },
                        canonical=True,
                    ),
                )

                self.hash_map[ns][digest_cnt] = hashfunc(
                    cbor2.dumps(self.disclosure_map[ns][digest_cnt], canonical=True)
                ).digest()

                digest_cnt += 1

    def format_datetime_repr(self, dt: datetime.datetime) -> str:
        """
        Format a datetime object to a string

        :param dt: datetime.datetime: the datetime object
        :return: str: the formatted string
        """
        return dt.isoformat().split(".")[0] + "Z"

    def sign(
        self,
        device_key: dict | None = None,
        valid_from: datetime.datetime | None = None,
        doctype: str | None = None,
    ) -> Sign1Message:
        """
        Sign a mso and returns it

        :param device_key: Union[dict, None]: the device key
        :param valid_from: Union[None, datetime.datetime]: the valid from date
        :param doctype: str: the document type

        :return: Sign1Message: the signed mso
        """

        utcnow = datetime.datetime.utcnow()
        valid_from = datetime.datetime.strptime(
            self.validity["issuance_date"], "%Y-%m-%d"
        )

        if settings.PYMDOC_EXP_DELTA_HOURS:
            exp = utcnow + datetime.timedelta(hours=settings.PYMDOC_EXP_DELTA_HOURS)
        else:
            # five years
            exp = datetime.datetime.strptime(self.validity["expiry_date"], "%Y-%m-%d")
            # exp = utcnow + datetime.timedelta(hours=(24 * 365) * 5)

        if utcnow > valid_from:
            valid_from = utcnow

        alg_map = {"ES256": "SHA-256", "ES384": "SHA-384", "ES512": "SHA-512"}

        payload = {
            "docType": doctype or list(self.hash_map)[0],
            "version": "1.0",
            "validityInfo": {
                "signed": cbor2.CBORTag(0, self.format_datetime_repr(utcnow)),
                "validFrom": cbor2.CBORTag(
                    0, self.format_datetime_repr(valid_from or utcnow)
                ),
                "validUntil": cbor2.CBORTag(0, self.format_datetime_repr(exp)),
            },
            "valueDigests": self.hash_map,
            "deviceKeyInfo": {
                "deviceKey": device_key,
            },
            "digestAlgorithm": alg_map.get(self.alg)
        }
        if self.revocation is not None:
            payload.update({"status": self.revocation})

        if self.cert_path:
            # Try to load the certificate file
            with open(self.cert_path, "rb") as file:
                certificate = file.read()
            _parsed_cert: Union[Certificate, None] = None
            try:
                _parsed_cert = x509.load_pem_x509_certificate(certificate)
            except Exception:
                logger.error(
                    f"Certificate at {self.cert_path} could not be loaded as PEM, trying DER"
                )

            if not _parsed_cert:
                try:
                    _parsed_cert = x509.load_der_x509_certificate(certificate)
                except Exception:
                    _err_msg = (
                        f"Certificate at {self.cert_path} could not be loaded as DER"
                    )
                    logger.error(_err_msg)

            if _parsed_cert:
                cert = _parsed_cert
            else:
                raise Exception(f"Certificate at {self.cert_path} failed parse")
            _cert = cert.public_bytes(getattr(serialization.Encoding, "DER"))
        else:
            if not self.cert_info:
                raise ValueError("cert_info must be provided if cert_path is not set")

            logger.warning(
                "A self-signed certificate will be created using the provided "
                "cert_info but this is not recommended for production use."
            )

            _cert = selfsigned_x509cert(self.cert_info, self.private_key)

        if self.hsm:
            # print("payload diganostic notation: \n",cbor2diag(cbor2.dumps(cbor2.CBORTag(24, cbor2.dumps(payload)))))

            mso = Sign1Message(
                phdr={
                    Algorithm: self.alg,
                    # 33: _cert
                },
                # TODO: x509 (cbor2.CBORTag(33)) and federation trust_chain support (cbor2.CBORTag(27?)) here
                # 33 means x509chain standing to rfc9360
                # in both protected and unprotected for interop purpose .. for now.
                uhdr={33: _cert},
                payload=cbor2.dumps(
                    cbor2.CBORTag(24, cbor2.dumps(payload, canonical=True)),
                    canonical=True,
                ),
            )

        else:
            logger.debug(
                "payload diagnostic notation: %s",
                cbor2diag(cbor2.dumps(cbor2.CBORTag(24, cbor2.dumps(payload)))),
            )

            mso = Sign1Message(
                phdr={
                    Algorithm: self.private_key.alg,
                    # KID: self.private_key.kid,
                    # 33: _cert
                },
                # TODO: x509 (cbor2.CBORTag(33)) and federation trust_chain support (cbor2.CBORTag(27?)) here
                # 33 means x509chain standing to rfc9360
                # in both protected and unprotected for interop purpose .. for now.
                uhdr={33: _cert},
                payload=cbor2.dumps(
                    cbor2.CBORTag(24, cbor2.dumps(payload, canonical=True)),
                    canonical=True,
                ),
            )

            mso.key = self.private_key

        return mso



import cbor2
from asn1crypto.x509 import Certificate
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.x509 import load_der_x509_certificate
from pycose.messages import Sign1Message

from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mdoc.verifier import MdocCbor
from pymdoccbor.mso.issuer import MsoIssuer
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.pid_data import PID_DATA
from pymdoccbor.tests.pkey import PKEY, PKEY_ED25519, PKEY_RSA


def extract_mso(mdoc: dict):
    mso_data = mdoc["documents"][0]["issuerSigned"]["issuerAuth"][2]
    mso_cbortag = cbor2.loads(mso_data)
    mso = cbor2.loads(mso_cbortag.value)
    return mso


def test_mso_writer():
    validity = {"issuance_date": "2025-01-17", "expiry_date": "2025-11-13"}
    msoi = MsoIssuer(
        data=PID_DATA,
        private_key=PKEY,
        validity=validity,
        alg="ES256",
        cert_info=CERT_DATA
    )

    assert "eu.europa.ec.eudiw.pid.1" in msoi.hash_map
    assert msoi.hash_map["eu.europa.ec.eudiw.pid.1"]

    assert "eu.europa.ec.eudiw.pid.1" in msoi.disclosure_map
    assert msoi.disclosure_map["eu.europa.ec.eudiw.pid.1"]
    assert msoi.disclosure_map["eu.europa.ec.eudiw.pid.1"].values().__len__() == PID_DATA["eu.europa.ec.eudiw.pid.1"].values().__len__()

    mso = msoi.sign()

    Sign1Message.decode(mso.encode())


def test_mdoc_issuer():
    validity = {"issuance_date": "2025-01-17", "expiry_date": "2025-11-13"}
    mdoci = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    with open("pymdoccbor/tests/certs/fake-cert.pem", "rb") as file:
        fake_cert_file = file.read()
        asl_signing_cert = x509.load_pem_x509_certificate(fake_cert_file)
        _asl_signing_cert = asl_signing_cert.public_bytes(getattr(serialization.Encoding, "DER"))
        status_list = {
            "status_list": {
                "idx": 0,
                "uri": "https://issuer.com/statuslists",
                "certificate": _asl_signing_cert,
            }
        }
        mdoc = mdoci.new(
            doctype="eu.europa.ec.eudiw.pid.1",
            data=PID_DATA,
            devicekeyinfo=PKEY,
            validity=validity,
            revocation=status_list
        )

    mdocp = MdocCbor()
    aa = cbor2.dumps(mdoc)
    mdocp.loads(aa)
    assert mdocp.verify() is True

    mdoci.dump()
    mdoci.dumps()

    # check mso content for status list
    mso = extract_mso(mdoc)
    status_list = mso["status"]["status_list"]
    assert status_list["idx"] == 0
    assert status_list["uri"] == "https://issuer.com/statuslists"
    cert_bytes = status_list["certificate"]
    cert: Certificate = load_der_x509_certificate(cert_bytes)
    assert "Test ASL Issuer" in cert.subject.rfc4514_string()


def test_mdoc_issuer_EdDSA():
    validity = {"issuance_date": "2025-01-17", "expiry_date": "2025-11-13"}
    mdoci = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    with open("pymdoccbor/tests/certs/fake-cert.pem", "rb") as file:
        fake_cert_file = file.read()
        asl_signing_cert = x509.load_pem_x509_certificate(fake_cert_file)
        _asl_signing_cert = asl_signing_cert.public_bytes(getattr(serialization.Encoding, "DER"))
        status_list = {
            "status_list": {
                "idx": 0,
                "uri": "https://issuer.com/statuslists",
                "certificate": _asl_signing_cert,
            }
        }
        mdoc = mdoci.new(
            doctype="eu.europa.ec.eudiw.pid.1",
            data=PID_DATA,
            devicekeyinfo=PKEY_ED25519,
            validity=validity,
            revocation=status_list
        )

    mdocp = MdocCbor()
    aa = cbor2.dumps(mdoc)
    mdocp.loads(aa)
    assert mdocp.verify() is True

    mdoci.dump()
    mdoci.dumps()

    # check mso content for status list
    mso = extract_mso(mdoc)
    status_list = mso["status"]["status_list"]
    assert status_list["idx"] == 0
    assert status_list["uri"] == "https://issuer.com/statuslists"
    cert_bytes = status_list["certificate"]
    cert: Certificate = load_der_x509_certificate(cert_bytes)
    assert "Test ASL Issuer" in cert.subject.rfc4514_string()


def test_mdoc_issuer_RSA():
    validity = {"issuance_date": "2025-01-17", "expiry_date": "2025-11-13"}
    mdoci = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
        cert_info=CERT_DATA
    )
    with open("pymdoccbor/tests/certs/fake-cert.pem", "rb") as file:
        fake_cert_file = file.read()
        asl_signing_cert = x509.load_pem_x509_certificate(fake_cert_file)
        _asl_signing_cert = asl_signing_cert.public_bytes(getattr(serialization.Encoding, "DER"))
        status_list = {
            "status_list": {
                "idx": 0,
                "uri": "https://issuer.com/statuslists",
                "certificate": _asl_signing_cert,
            }
        }
        mdoc = mdoci.new(
            doctype="eu.europa.ec.eudiw.pid.1",
            data=PID_DATA,
            devicekeyinfo=PKEY_RSA,
            validity=validity,
            revocation=status_list
        )

    mdocp = MdocCbor()
    aa = cbor2.dumps(mdoc)
    mdocp.loads(aa)
    assert mdocp.verify() is True

    mdoci.dump()
    mdoci.dumps()

    # check mso content for status list
    mso = extract_mso(mdoc)
    status_list = mso["status"]["status_list"]
    assert status_list["idx"] == 0
    assert status_list["uri"] == "https://issuer.com/statuslists"
    cert_bytes = status_list["certificate"]
    cert: Certificate = load_der_x509_certificate(cert_bytes)
    assert "Test ASL Issuer" in cert.subject.rfc4514_string()



import base64
import binascii
import logging
from datetime import datetime, timezone
from typing import Union

import cbor2
from cbor_diag import cbor2diag
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from pycose.keys import CoseKey, EC2Key

from pymdoccbor.mdoc.exceptions import InvalidStatusDescriptor
from pymdoccbor.mso.issuer import MsoIssuer

logger = logging.getLogger("pymdoccbor")


class MdocCborIssuer:
    """
    MdocCborIssuer helper class to create a new mdoc
    """

    def __init__(
        self,
        key_label: str | None = None,
        user_pin: str | None = None,
        lib_path: str | None = None,
        slot_id: int | None = None,
        hsm: bool = False,
        alg: str | None = None,
        kid: str | None = None,
        private_key: Union[dict, CoseKey] = {},
        cert_info: dict | None = None,
    ):
        """
        Initialize a new MdocCborIssuer

        :param key_label: str: key label
        :param user_pin: str: user pin
        :param lib_path: str: path to the library cryptographic library
        :param slot_id: int: slot id
        :param hsm: bool: hardware security module
        :param alg: str: hashig algorithm
        :param kid: str: key id
        :param private_key: Union[dict, CoseKey]: private key
        """
        self.version: str = "1.0"
        self.status: int = 0

        if private_key:
            if isinstance(private_key, dict):
                self.private_key = CoseKey.from_dict(private_key)
            elif isinstance(private_key, EC2Key):
                ec2_encoded = private_key.encode()
                ec2_decoded = CoseKey.decode(ec2_encoded)
                self.private_key = ec2_decoded
            elif isinstance(private_key, CoseKey):
                self.private_key = private_key
            else:
                raise ValueError("private_key must be a dict or CoseKey object")

        self.signed: dict = {}
        self.key_label = key_label
        self.user_pin = user_pin
        self.lib_path = lib_path
        self.slot_id = slot_id
        self.hsm = hsm
        self.alg = alg
        self.kid = kid
        self.cert_info = cert_info

    def new(
        self,
        data: dict,
        doctype: str,
        validity: dict | None = None,
        devicekeyinfo: dict | CoseKey | str | None = None,
        cert_path: str | None = None,
        revocation: dict | None = None,
        status: dict | None = None
    ) -> dict:
        """
        create a new mdoc with signed mso

        :param data: dict: data to be signed
        :param doctype: str: document type
        :param validity: dict: validity info
        :param devicekeyinfo: Union[dict, CoseKey, str]: device key info
        :param cert_path: str: path to the certificate
        :param revocation: dict: revocation status dict it may include status_list and identifier_list keys
        :param status: dict: status dict with uri and idx per draft-ietf-oauth-status-list
        :return: dict: signed mdoc
        """
        if isinstance(devicekeyinfo, dict):
            devicekeyinfoCoseKeyObject = CoseKey.from_dict(devicekeyinfo)
            if devicekeyinfoCoseKeyObject.kty.identifier == 2:  # EC2Key
                devicekeyinfo = {
                    1: devicekeyinfoCoseKeyObject.kty.identifier,
                    -1: devicekeyinfoCoseKeyObject.crv.identifier,
                    -2: devicekeyinfoCoseKeyObject.x,
                    -3: devicekeyinfoCoseKeyObject.y,
                }
            elif devicekeyinfoCoseKeyObject.kty.identifier == 1:  # OKPKey
                devicekeyinfo = {
                    1: devicekeyinfoCoseKeyObject.kty.identifier,
                    -1: devicekeyinfoCoseKeyObject.crv.identifier,
                    -2: devicekeyinfoCoseKeyObject.x,
                }
            elif devicekeyinfoCoseKeyObject.kty.identifier == 3:  # RSAKey
                devicekeyinfo = {
                    1: devicekeyinfoCoseKeyObject.kty.identifier,
                    -1: devicekeyinfoCoseKeyObject.n,
                    -2: devicekeyinfoCoseKeyObject.e,
                }
            else:
                raise TypeError("Unsupported key type in devicekeyinfo")
        if isinstance(devicekeyinfo, str):
            device_key_bytes = base64.urlsafe_b64decode(devicekeyinfo.encode("utf-8"))
            public_key = serialization.load_pem_public_key(device_key_bytes)

            if isinstance(public_key, EllipticCurvePublicKey):
                curve_name = public_key.curve.name
                curve_map = {
                    "secp256r1": 1,  # NIST P-256
                    "secp384r1": 2,  # NIST P-384
                    "secp521r1": 3,  # NIST P-521
                    "brainpoolP256r1": 8,  # Brainpool P-256
                    "brainpoolP384r1": 9,  # Brainpool P-384
                    "brainpoolP512r1": 10,  # Brainpool P-512
                    # Add more curve mappings as needed
                }
                curve_identifier = curve_map.get(curve_name)

                # Extract the x and y coordinates from the public key
                x = public_key.public_numbers().x.to_bytes(
                    (public_key.public_numbers().x.bit_length() + 7)
                    // 8,  # Number of bytes needed
                    "big",  # Byte order
                )

                y = public_key.public_numbers().y.to_bytes(
                    (public_key.public_numbers().y.bit_length() + 7)
                    // 8,  # Number of bytes needed
                    "big",  # Byte order
                )

                devicekeyinfo = {
                    1: 2,
                    -1: curve_identifier,
                    -2: x,
                    -3: y,
                }
            elif isinstance(public_key, Ed25519PublicKey):
                devicekeyinfo = {
                    1: 1,  # OKPKey
                    -1: "Ed25519",  # Curve identifier for Ed25519
                    -2: public_key.public_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PublicFormat.Raw
                    ),
                }
            elif isinstance(public_key, RSAPublicKey):
                devicekeyinfo = {
                    1: 3,  # RSAKey
                    -1: public_key.public_numbers().n.to_bytes(
                        (public_key.public_numbers().n.bit_length() + 7) // 8, "big"
                    ),
                    -2: public_key.public_numbers().e.to_bytes(
                        (public_key.public_numbers().e.bit_length() + 7) // 8, "big"
                    )
                }
            else:
                raise TypeError("Loaded public key is not an EllipticCurvePublicKey")

        if self.hsm:
            msoi = MsoIssuer(
                data=data,
                cert_path=cert_path,
                hsm=self.hsm,
                key_label=self.key_label,
                user_pin=self.user_pin,
                lib_path=self.lib_path,
                slot_id=self.slot_id,
                alg=self.alg,
                kid=self.kid,
                validity=validity,
                revocation=revocation,
                cert_info=self.cert_info
            )

        else:
            msoi = MsoIssuer(
                data=data,
                private_key=self.private_key,
                alg=self.alg,
                cert_path=cert_path,
                validity=validity,
                revocation=revocation,
                cert_info=self.cert_info
            )

        mso = msoi.sign(
            doctype=doctype,
            device_key=devicekeyinfo,
            valid_from=datetime.now(timezone.utc)
        )

        mso_cbor = mso.encode(
            tag=False,
            hsm=self.hsm,
            key_label=self.key_label,
            user_pin=self.user_pin,
            lib_path=self.lib_path,
            slot_id=self.slot_id,
        )

        res = {
            "version": self.version,
            "documents": [
                {
                    "docType": doctype,  # 'org.iso.18013.5.1.mDL'
                    "issuerSigned": {
                        "nameSpaces": {
                            ns: [v for k, v in dgst.items()]
                            for ns, dgst in msoi.disclosure_map.items()
                        },
                        "issuerAuth": cbor2.loads(mso_cbor)
                    },
                }
            ],
            "status": self.status,
        }

        if status:
            if "status_list" not in status:
                raise InvalidStatusDescriptor("status_list is required")

            if "uri" not in status["status_list"]:
                raise InvalidStatusDescriptor("uri is required")
            if "idx" not in status["status_list"]:
                raise InvalidStatusDescriptor("idx is required")

            res["status"] = status

        logger.debug(f"MSO diagnostic notation: {cbor2diag(mso_cbor)}")

        self.signed = res
        return self.signed

    def dump(self) -> bytes:
        """
        Returns the CBOR representation of the signed mdoc

        :return: bytes: CBOR representation of the signed mdoc
        """
        return cbor2.dumps(self.signed, canonical=True)

    def dumps(self) -> bytes:
        """
        Returns the AF binary representation of the signed mdoc

        :return: bytes: AF binary representation of the signed mdoc
        """
        return binascii.hexlify(cbor2.dumps(self.signed, canonical=True))



PID_DATA = {
    "eu.europa.ec.eudiw.pid.1": {
        "family_name": "Raffaello",
        "given_name": "Mascetti",
        "birth_date": "1922-03-13",
        "age_over_18": True,
        "age_over_NN": 150,
        "age_in_years": 101,
        "age_birth_year": 1922,
        "unique_id": "random-opaque-value",
        "family_name_birth": "Mascetti",
        "given_name_birth": "Raffaello",
        "birth_place": "Firenze",
        "birth_country": "IT",
        "birth_state": "Tuscany",
        "birth_city": "Firenze",
        "resident_address": "via Tal dei tali",
        "resident_country": "Italy",
        "resident_state": "Tuscany",
        "resident_city": "Firenze",
        "resident_postal_code": "50100",
        "resident_street": "via Tal dei tali",
        "resident_house_number": "21",
        "gender": 1,
        "nationality": "IT",
        "portrait": None,
        "portrait_capture_date": None,
        "issuance_date": "2023-06-07",
        "expiry_date": "2027-06-07",
        "issuing_authority": "IT",
        "document_number": "CA0568IT_example",
        "administrative_number": None,
        "issuing_country": "IT",
        "issuing_jurisdiction": None
    }
}



import re
from setuptools import setup, find_packages


def readme():
    with open("README.md") as f:
        return f.read()


_pkg_name = "pymdoccbor"

with open(f"{_pkg_name}/__init__.py", "r") as fd:
    VERSION = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

setup(
    name=_pkg_name,
    version=VERSION,
    description="Python parser and writer for Mobile Driving License and EUDI Wallet MDOC CBOR.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    url="https://github.com/IdentityPython/pyMDL-MDOC",
    author="Giuseppe De Marco",
    author_email="demarcog83@gmail.com",
    license="License :: OSI Approved :: Apache Software License",
    # scripts=[f'{_pkg_name}/bin/{_pkg_name}'],
    packages=find_packages(include=["pymdoccbor", "pymdoccbor.*"]),
    include_package_data=True,
    install_requires=["cbor2>=5.4.0", "cbor-diag>=1.1.0", "pycose>=1.0.1"],
)



from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY

mdoc = MdocCborIssuer(
    private_key=PKEY,
    alg="ES256",
    cert_info=CERT_DATA
)


def test_MdocCborIssuer_creation():
    assert mdoc.version == '1.0'
    assert mdoc.status == 0


def test_mdoc_without_private_key_must_fail():
    try:
        MdocCborIssuer(None)
    except Exception as e:
        assert str(e) == "You must provide a private key"


def test_MdocCborIssuer_new_single():
    mdoc.new(
        data=MICOV_DATA,
        # devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )
    assert mdoc.signed['version'] == '1.0'
    assert mdoc.signed['status'] == 0
    assert mdoc.signed['documents'][0]['docType'] == 'org.micov.medical.1'
    assert mdoc.signed['documents'][0]['issuerSigned']['nameSpaces']['org.micov.medical.1'][0].tag == 24

# TODO: restore multiple documents support
# def test_MdocCborIssuer_new_multiple():
#     micov_data = {"doctype": "org.micov.medical.1", "data": MICOV_DATA}
#     pid_data = {"doctype": "eu.europa.ec.eudiw.pid.1", "data": PID_DATA}
#     mdoc.new(
#         doctype="org.micov.medical.1",
#         data=[micov_data, pid_data],
#         validity={"issuance_date": "2024-12-31", "expiry_date": "2050-12-31"},
#         # devicekeyinfo=PKEY  # TODO
#     )
#     assert mdoc.signed['version'] == '1.0'
#     assert mdoc.signed['documents'][0]['docType'] == 'org.micov.medical.1'
#     assert mdoc.signed['documents'][1]['docType'] == 'eu.europa.ec.eudiw.pid.1'


def test_MdocCborIssuer_dump():
    dump = mdoc.dump()

    assert dump
    assert isinstance(dump, bytes)
    assert len(dump) > 0


def test_MdocCborIssuer_dumps():
    dumps = mdoc.dumps()

    assert dumps
    assert isinstance(dumps, bytes)
    assert len(dumps) > 0



import logging
from typing import Union

import cbor2
import cryptography
from cryptography.exceptions import InvalidSignature
from pycose.keys import EC2Key
from pycose.messages import Sign1Message

from pymdoccbor import settings
from pymdoccbor.exceptions import (MsoX509ChainNotFound,
                                   UnsupportedMsoDataFormat)
from pymdoccbor.tools import bytes2CoseSign1, cborlist2CoseSign1

logger = logging.getLogger("pymdoccbor")


class MsoVerifier:
    """
    Parameters
        data: CBOR TAG 24

    Example:
        MsoParser(mdoc['documents'][0]['issuerSigned']['issuerAuth'])

    Note
        The signature is contained in an untagged COSE_Sign1
        structure as defined in RFC 8152.
    """

    def __init__(self, data: Union[cbor2.CBORTag, bytes, list]) -> None:
        """
        Initialize the MsoParser object

        :param data: Union[cbor2.CBORTag, bytes, list]: the data to parse
        """

        self._data = data

        # not used
        if isinstance(self._data, bytes):
            self.object: Sign1Message = bytes2CoseSign1(
                cbor2.dumps(cbor2.CBORTag(18, value=self._data)))
        elif isinstance(self._data, list):
            self.object: Sign1Message = cborlist2CoseSign1(self._data)
        else:
            raise UnsupportedMsoDataFormat(
                f"MsoParser only supports raw bytes and list, a {type(data)} was provided"
            )

        self.object.key = None
        self.public_key: cryptography.hazmat.backends.openssl.ec._EllipticCurvePublicKey = None
        self.x509_certificates: list = []
        self.verified_root = None  # Will store the trusted root that verified the chain

    @property
    def payload_as_cbor(self) -> dict:
        """
        It returns the payload as a CBOR TAG

        :return: dict: the payload as a CBOR TAG 24
        """
        return cbor2.loads(self.object.payload)

    @property
    def payload_as_raw(self):
        return self.object.payload

    @property
    def payload_as_dict(self):
        return cbor2.loads(
            cbor2.loads(self.object.payload).value
        )

    @property
    def raw_public_keys(self) -> list[Union[bytes, dict]]:
        """
            Extracts public keys from x509 certificates found in the MSO.
            This method searches for x509 certificates in both the protected header (phdr)
            and unprotected header (uhdr) of the COSE_Sign1 object. It handles certificate
            data in various formats, including:
            - `bytes`: Returns a list containing the raw bytes of the certificate.
            - `list`: Returns the list of certificates as-is.
            - `dict`: Wraps the dictionary in a list and returns it.
            If no valid x509 certificates are found, an `MsoX509ChainNotFound` exception
            is raised. Unexpected types are logged as warnings.
            :return: list[Any]: A list of certificates in their respective formats.
            :raises MsoX509ChainNotFound: If no x509 certificates are found.
        """
        merged = self.object.phdr.copy()
        merged.update(self.object.uhdr)
        _mixed_heads = merged.items()
        for h, v in _mixed_heads:
            if h.identifier == 33:
                if isinstance(v, bytes):
                    return [v]
                elif isinstance(v, list):
                    return v
                elif isinstance(v, dict):
                    return [v]
                else:
                    logger.warning(
                        f"Unexpected type for public key: {type(v)}. "
                        "Expected bytes, list or dict."
                    )
                    continue

        raise MsoX509ChainNotFound(
            "I can't find any valid X509certs, identified by label number 33, "
            "in this MSO."
        )

    def attest_public_key(self, trusted_root_certs: list = None):
        """
        Verify the X.509 certificate chain.

        Args:
            trusted_root_certs: List of trusted root certificates (x509.Certificate objects)
                               If None, skips chain validation (backward compatible)

        Returns:
            The trusted root certificate that signed the DS cert, or None if validation skipped
        """
        if trusted_root_certs is None:
            logger.warning(
                "Certificate chain validation skipped. "
                "Pass trusted_root_certs parameter to verify() to enable X.509 chain validation."
            )
            return None

        # Load DS certificate (first in chain)
        ds_cert = self.x509_certificates[0] if self.x509_certificates else None
        if not ds_cert:
            raise ValueError("No DS certificate found in MSO")

        # Verify DS cert is signed by one of the trusted roots
        verified_root = None
        for root_cert in trusted_root_certs:
            try:
                # Verify signature
                root_cert.public_key().verify(
                    ds_cert.signature,
                    ds_cert.tbs_certificate_bytes,
                    ds_cert.signature_algorithm_parameters
                )
                verified_root = root_cert
                logger.info(f"Certificate chain verified with root: {root_cert.subject}")
                break
            except InvalidSignature:
                continue
            except Exception as exc:
                logger.warning(f"Error verifying with root cert: {exc}")
                continue

        if not verified_root:
            raise ValueError("DS certificate not signed by any trusted root")

        # Verify certificate validity dates
        from datetime import datetime, timezone
        now = datetime.now(timezone.utc)

        if ds_cert.not_valid_before_utc > now:
            raise ValueError(f"DS certificate not yet valid (valid from {ds_cert.not_valid_before_utc})")

        if ds_cert.not_valid_after_utc < now:
            raise ValueError(
                f"DS certificate expired (valid until {ds_cert.not_valid_after_utc})"
            )

        logger.info("Certificate chain and validity verified successfully")
        return verified_root

    def load_public_key(self, trusted_root_certs: list = None) -> None:
        """
        Load the public key from the x509 certificate

        Args:
            trusted_root_certs: List of trusted root certificates for chain validation
        :return: None
        """
        for i in self.raw_public_keys:
            self.x509_certificates.append(
                cryptography.x509.load_der_x509_certificate(i)
            )

        self.verified_root = self.attest_public_key(trusted_root_certs)

        self.public_key = self.x509_certificates[0].public_key()

        key = EC2Key(
            crv=settings.COSEKEY_HAZMAT_CRV_MAP[self.public_key.curve.name],
            x=self.public_key.public_numbers().x.to_bytes(
                settings.CRV_LEN_MAP[self.public_key.curve.name], 'big'
            ),
            y=self.public_key.public_numbers().y.to_bytes(
                settings.CRV_LEN_MAP[self.public_key.curve.name], 'big'
            )
        )
        self.object.key = key

    def verify_signature(self, trusted_root_certs: list = None) -> bool:
        """"
        Verify the signature of the MSO

        Args:
            trusted_root_certs: List of trusted root certificates for chain validation
        :return: bool: True if the signature is valid, False otherwise
        """
        if not self.object.key:
            self.load_public_key(trusted_root_certs)

        return self.object.verify_signature()

    def verify_element_hashes(self, namespaces: dict) -> dict:
        """
        Verify that disclosed elements match their hashes in the MSO.

        Args:
            namespaces: The nameSpaces dict from IssuerSigned containing IssuerSignedItems

        Returns:
            dict: Results with 'valid' (bool), 'total' (int), 'verified' (int), 'failed' (list)
        """
        import hashlib

        mso_data = self.payload_as_dict
        value_digests = mso_data.get('valueDigests', {})

        results = {
            'valid': True,
            'total': 0,
            'verified': 0,
            'failed': []
        }

        for namespace, items in namespaces.items():
            if namespace not in value_digests:
                logger.warning(f"Namespace {namespace} not found in MSO valueDigests")
                continue

            namespace_digests = value_digests[namespace]

            for item_bytes in items:
                results['total'] += 1

                # item_bytes might be a CBORTag object, need to encode it
                if isinstance(item_bytes, cbor2.CBORTag):
                    item_bytes_raw = cbor2.dumps(item_bytes)
                else:
                    item_bytes_raw = item_bytes

                # Decode to get digestID
                try:
                    item_data = cbor2.loads(item_bytes_raw)
                    if isinstance(item_data, cbor2.CBORTag) and item_data.tag == 24:
                        item_content = cbor2.loads(item_data.value)
                    else:
                        item_content = item_data

                    digest_id = item_content.get('digestID')
                    element_id = item_content.get('elementIdentifier')

                    # Compute hash of the full tagged bytes
                    computed_hash = hashlib.sha256(item_bytes_raw).digest()

                    # Get expected hash from MSO
                    expected_hash = namespace_digests.get(digest_id)

                    if expected_hash is None:
                        logger.error(f"digestID {digest_id} not found in MSO for {namespace}/{element_id}")
                        results['failed'].append({
                            'namespace': namespace,
                            'digestID': digest_id,
                            'elementIdentifier': element_id,
                            'reason': 'digestID not in MSO'
                        })
                        results['valid'] = False
                        continue

                    if computed_hash != expected_hash:
                        logger.error(f"Hash mismatch for {namespace}/{element_id} (digestID={digest_id})")
                        results['failed'].append({
                            'namespace': namespace,
                            'digestID': digest_id,
                            'elementIdentifier': element_id,
                            'reason': 'hash mismatch',
                            'expected': expected_hash.hex(),
                            'computed': computed_hash.hex()
                        })
                        results['valid'] = False
                    else:
                        results['verified'] += 1
                        logger.debug(f"Hash verified for {namespace}/{element_id}")

                except Exception as e:
                    logger.error(f"Error verifying element hash: {e}")
                    results['failed'].append({
                        'namespace': namespace,
                        'reason': f'exception: {e}'
                    })
                    results['valid'] = False

        return results



import datetime

import cbor2

from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.mdoc.verifier import MdocCbor
from pymdoccbor.tests.cert_data import CERT_DATA
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.tests.pkey import PKEY


def normalize_dates(obj):
    if isinstance(obj, dict):
        return {k: normalize_dates(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [normalize_dates(v) for v in obj]
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    return obj


def test_mdoc_cbor_creation():
    mdoci = MdocCborIssuer(private_key=PKEY, alg="ES256", cert_info=CERT_DATA)
    mdoc = mdoci.new(
        data=MICOV_DATA,
        # devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={"issuance_date": "2024-12-31", "expiry_date": "2050-12-31"},
        status={
            "status_list": {"idx": 412, "uri": "https://example.com/statuslists/1"}
        },
    )

    data = cbor2.dumps(mdoc, datetime_as_timestamp=True)

    mdocp = MdocCbor()
    mdocp.loads(data)
    mdocp.verify()

    assert mdoc
    assert "org.micov.medical.1" in mdocp.disclosure_map

    assert normalize_dates(mdocp.disclosure_map) == normalize_dates(MICOV_DATA)
    assert mdocp.status == {
        "status_list": {"idx": 412, "uri": "https://example.com/statuslists/1"}
    }


def test_mdoc_cbor_invalid_status():
    mdoci = MdocCborIssuer(private_key=PKEY, alg="ES256", cert_info=CERT_DATA)

    try:
        mdoci.new(
            data=MICOV_DATA,
            # devicekeyinfo=PKEY,  # TODO
            doctype="org.micov.medical.1",
            validity={"issuance_date": "2024-12-31", "expiry_date": "2050-12-31"},
            status={
                "status_list": {
                    "idx": 412,
                    # "uri": "https://example.com/statuslists/1"  # Missing URI
                }
            },
        )
    except Exception as e:
        assert str(e) == "uri is required"

    try:
        mdoci.new(
            data=MICOV_DATA,
            # devicekeyinfo=PKEY,  # TODO
            doctype="org.micov.medical.1",
            validity={"issuance_date": "2024-12-31", "expiry_date": "2050-12-31"},
            status={
                "status_list": {
                    # "idx": 412,
                    "uri": "https://example.com/statuslists/1"  # Missing URI
                }
            },
        )
    except Exception as e:
        assert str(e) == "idx is required"

    try:
        mdoci.new(
            data=MICOV_DATA,
            # devicekeyinfo=PKEY,  # TODO
            doctype="org.micov.medical.1",
            validity={"issuance_date": "2024-12-31", "expiry_date": "2050-12-31"},
            status={
                "not_status_list": {
                    "idx": 412,
                    "uri": "https://example.com/statuslists/1",  # Missing URI
                }
            },
        )
    except Exception as e:
        assert str(e) == "status_list is required"



import os

COSEKEY_HAZMAT_CRV_MAP = {
    "secp256r1": "P_256",
    "secp384r1": "P_384",
    "secp521r1":  "P_521"
}

CRV_LEN_MAP = {
    "secp256r1": 32,
}

PYMDOC_HASHALG: str = os.getenv("PYMDOC_HASHALG", "SHA-256")
PYMDOC_EXP_DELTA_HOURS: int = os.getenv("PYMDOC_EXP_DELTA_HOURS", 0)

HASHALG_MAP = {
    "SHA-256": "sha256",
    "SHA-512": "sha512",
}

DIGEST_SALT_LENGTH = 32

X509_DER_CERT = os.getenv("X509_DER_CERT", None)

CBORTAGS_ATTR_MAP = {
    "birth_date": 1004,
    "expiry_date": 1004,
    "issue_date": 1004,
    "issuance_date": 1004,
}
